package transfer

// The "transfer" package is the main NSTA running on the master server

import (
	"context"
	"fmt"
	"io"
	"log"
	"os"
	"os/signal"
	"sync"
	"syscall"
	"time"

	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpserver"
	"github.com/it-novum/nsta/oitcdb"
	"github.com/it-novum/nsta/satstatus"
	"github.com/it-novum/nsta/sshwrapper"
)

type SatelliteOptions struct {
	//all
	StatusTracker *satstatus.StatusTracker
	DB            *oitcdb.Satellite

	NagiosPath string

	ToMaster chan<- *gearman.Message

	// ssh
	SSHConnOpts         sshwrapper.ConnOpts
	SSHReloadTimeoutSec int

	// http_push
	RegisterSatelliteIO httpserver.RegisterSatelliteIO
}

// Satellite is the common interface of all different satellite types
type Satellite interface {
	SatelliteID() int
	SatelliteName() string
	Start(ctx context.Context) error
	Close()
	Compare(*oitcdb.Satellite) bool
	SendMessage(msg *gearman.Message)
	Reload(ctx context.Context)
}

// closeSatellites is a helper that closes all satellites
func closeSatellites(satellites []Satellite) {
	if len(satellites) > 0 {
		var wg sync.WaitGroup
		wg.Add(len(satellites))
		for _, sat := range satellites {
			go func(sat Satellite) {
				sat.Close()
				wg.Done()
			}(sat)
		}
		wg.Wait()
	}
}

// MainSystem handles the master to satellite connections and incoming and outgoing messages
type MainSystem struct {
	AccessLogWriter       io.Writer
	MyCnf                 string
	GearmanAddress        string
	WebserverAddressHTTP  string
	WebserverAddressHTTPS string
	CertFile              string
	KeyFile               string
	NagiosPath            string
	SSHReloadTimeoutSec   int

	oitc *oitcdb.OITC

	webserver     *httpserver.NSTAHTTPServer
	localGearman  *gearman.Client
	satellites    map[int]Satellite
	toLocal       gearman.SyncMessageFIFO
	toSatFan      gearman.SyncMessageFIFO
	statusTracker *satstatus.StatusTracker
}

// Run starts the actual work here
func (m *MainSystem) run(ctx context.Context) {
	// Bind signal handler for linux kernel signals
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM, syscall.SIGHUP)

	m.localGearman.Start(ctx)
	m.reload(ctx) // Load config and start Satellite instances
	m.webserver.Run()

	defer func() {
		m.webserver.Stop()
		sats := make([]Satellite, 0)
		for _, sat := range m.satellites {
			sats = append(sats, sat)
		}
		//Keep the order !
		closeSatellites(sats)
		m.toLocal.Close()
		m.toSatFan.Close()
		m.localGearman.Stop()
		m.statusTracker.Close()
		m.oitc.Close()
	}()

	for {
		select {
		case <-ctx.Done():
			return
		case sig := <-sigs:
			// Catch Kernel signals for reload or stop
			log.Println(&Error{Name: "Master", Message: fmt.Sprint("received signal:", sig.String())})
			if sig.String() == "hangup" {
				// reload config and re-start Satellite instances
				m.reload(ctx)
			} else {
				// Stop NSTA - break endless loop
				return
			}
		case <-m.toSatFan.Notify():
			// Cache notfiy signal from FIFO and send message to Satellite System
			msg := m.toSatFan.Get()
			if msg != nil {
				satID := msg.Content().SatelliteID
				sat, ok := m.satellites[satID]
				if ok {
					sat.SendMessage(msg)
				} else {
					log.Println(&Error{Name: "Master", Message: fmt.Sprint("unknown satellite id: ", satID)})
				}
			}
		case <-m.toLocal.Notify():
			// Catch notify signal from FIFO chache and pass data to the Statusengine Broker via Gearman
			msgs := m.toLocal.GetAll(8192)
			for _, msg := range msgs {
				m.localGearman.SendMessage(msg)
			}
		}
	}
}

func (m *MainSystem) reload(ctx context.Context) {
	log.Println(&Error{Name: "Local", Message: "reload of satellite configuration started"})
	dbSats := m.oitc.FetchSatellites()

	// Diff old configuration with new one to only restart the satellite instances
	// where the configuration (ipaddress, SSH to HTTP, etc) has changed
	newSatMap := make(map[int]Satellite)
	keepSats := make([]Satellite, 0)
	dropSats := make([]Satellite, 0)
	newDbSats := make([]*oitcdb.Satellite, 0)

	for satID, sat := range m.satellites {
		keep := false
	dbloop:
		for _, dbSat := range dbSats {
			if sat.Compare(dbSat) {
				keep = true
				break dbloop
			}
		}
		if keep {
			// Same config - nothing to do
			keepSats = append(keepSats, sat)
			newSatMap[satID] = sat
			log.Println(&Error{Name: sat.SatelliteName(), Message: "satellite configuration not changed -> keep running"})
		} else {
			// Changed configuration or deleted satellite - drop it
			log.Println(&Error{Name: sat.SatelliteName(), Message: "satellite configuration changed or deleted -> close"})
			dropSats = append(dropSats, sat)
		}
	}

	for _, dbSat := range dbSats {
		exists := false
	keeploop:
		for _, sat := range keepSats {
			if sat.Compare(dbSat) {
				exists = true
				break keeploop
			}
		}
		if !exists {
			// Create a list of all changed or new satellites
			log.Println(&Error{Name: dbSat.Name, Message: "satellite configuration new -> start"})
			newDbSats = append(newDbSats, dbSat)
		}
	}

	beginReload := time.Now()

	//Close all old satellites
	closeSatellites(dropSats)

	// Create new satellite instances
	for _, dbSat := range newDbSats {
		opts := SatelliteOptions{
			StatusTracker: m.statusTracker,
			DB:            dbSat,
			NagiosPath:    m.NagiosPath,
			ToMaster:      m.toLocal.Direct(), // Sends result from Satellite into FIFO cache (FIFO signals notify chan)
			SSHConnOpts: sshwrapper.ConnOpts{
				Hostname:       dbSat.Address,
				Login:          dbSat.Login,
				Port:           dbSat.Port,
				PrivateKeyPath: dbSat.PrivateKeyPath,
				RemotePort:     dbSat.RemotePort,
			},
			RegisterSatelliteIO: m.webserver,
			SSHReloadTimeoutSec: m.SSHReloadTimeoutSec,
		}
		var sat Satellite
		if dbSat.SyncMethod == "ssh" {
			sat = &SSHSatellite{
				SatelliteOptions: opts,
			}
		} else if dbSat.SyncMethod == "https_pull" {
			sat = &PullSatellite{
				SatelliteOptions: opts,
			}
		} else if dbSat.SyncMethod == "https_push" {
			sat = &PushSatellite{
				SatelliteOptions: opts,
			}
		} else {
			log.Println(&Error{Name: dbSat.Name, Message: fmt.Sprint("unknown SyncMethod:", dbSat.SyncMethod)})
			continue
		}

		// Start satellite instance / thread
		if err := sat.Start(ctx); err != nil {
			log.Println(&Error{Name: dbSat.Name, Message: fmt.Sprint("could not create", dbSat.SyncMethod, "server"), Err: err})
		} else {
			newSatMap[dbSat.ID] = sat
		}
	}

	// Insert satellite system status into database
	satList := make([]satstatus.Satellite, 0)
	for _, sat := range newSatMap {
		satList = append(satList, sat)
	}
	m.statusTracker.Reload(satList, beginReload)

	// Trigger reload for each satellite in parallel
	// And notify the satellite instances to copy the new monitoring configuration files
	wg := sync.WaitGroup{}
	wg.Add(len(newSatMap))
	for _, sat := range newSatMap {
		go func(ls Satellite) {
			ls.Reload(ctx)
			wg.Done()
		}(sat)
	}
	wg.Wait()

	m.satellites = newSatMap
	log.Println(&Error{Name: "Local", Message: "reload of satellite configuration finished"})
}

func (m *MainSystem) StartForeground() error {
	if oitc, err := oitcdb.New(m.MyCnf); err != nil {
		return err
	} else {
		m.oitc = oitc
	}

	m.webserver = httpserver.New(m.AccessLogWriter, m.NagiosPath, m.WebserverAddressHTTP, m.WebserverAddressHTTPS, m.CertFile, m.KeyFile, httpserver.WebserverTypeMaster)
	m.statusTracker = satstatus.New(m.oitc)
	m.toLocal.Start()
	m.toSatFan.Start()

	m.localGearman = &gearman.Client{
		ClientOptions: gearman.ClientOptions{
			ClientType:     gearman.Main,
			ClientName:     "Local",
			Address:        m.GearmanAddress,
			MessageReceive: m.toSatFan.Direct(), // Sends messages from Master System into FIFO cache (to send it to the Satellite Systems)
		},
	}

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()
	m.run(ctx)

	return nil
}
